/*
  UIDMP.CPP
  Main source file for class UIDmp and some misc. functions


  Copyright (c) 1994 by Borland International
  All Rights Reserved

  This source code is hereby released into the public domain.
  Written by Jeffrey J. Peters

*/

#include <dos.h>
#include <ctype.h>
#include <stdio.h>
#include "uiclass.h"

UIDmp::UIDmp (UIWindow &p, bits16 sel, bits32 ofs) : parent (p), curSel (sel),
              curOfs (ofs), dmpWidth (8), des (*new TDescriptor (sel)),
              srchSel(0), srchOfs(0), spin (*this),
              UIWindow (UIRect(p.rFrame.ul.x+2,p.rFrame.ul.y+2,
                                      p.rFrame.ul.x+55,p.rFrame.ul.y+9)," Dump ")
{

  uiFlags |= UIFNoScroll;
  uiFlags |= UIFNoCursorFix;
  if (p.rFrame.ul.y > 12)
    MoveTo (UIRect (rFrame)+= UIRect (0,-8,0,-8));

  DoNewSelector (sel);
}
#pragma argsused
void UIDmp::DoNewSelector (bits16 s)
{
#ifdef __DPMI32__
  des.Free ();
  if ((s == _CS) || (s == _DS))
    ;
  else
  {
    des.Allocate ();
    des.SetBase (des.base);
    des.SetLimit (des.limit);
    des.SetRights (des.rights&(~0x0F));  // force to be data/read/up/!acc
  }
#endif
  char *buf = new char [40];
  sprintf (buf, " Dump of Selector %04X ", curSel);
  if (title)
    delete title;
  title = buf;
}

bits08 GetRights (bits16 s)
{
#ifdef __DPMI32__
  PUSH_BX
  _EBX = 0;
#endif
  _BX = s;
  LAR_EAX_EBX;
#ifdef __DPMI32__
  POP_BX
#endif
  return _AH;
}

#ifdef __DPMI32__
bits08 Get32Rights (bits16 s)
{
  PUSH_BX
  _EBX = 0;
  _BX = s;
  LAR_EAX_EBX;
  POP_BX
  SHL_EAX_16;
  return _AH;
}
#endif

bits32 GetBase (bits16 s)
{
  bits16 upr, lwr;

#ifdef __DPMI32__
  PUSH_BX
  _EBX = 0;
  _EAX = 0;
#endif
  _BX = s;
  _AX = 6;
  geninterrupt (0x31);
#ifdef __DPMI32__
  POP_BX
#endif
  if (_FLAGS&1)
  {
    return -1;
  }
  upr = _CX;
  lwr = _DX;
  return ((bits32)upr)*0x10000L+lwr;
}

#pragma warn -rlv
bits32 GetLimit (bits16 s)
{
  if (!IsValidSel (s))
    return 0;

#if defined (__DPMI32__)
  PUSH_BX
  _EBX = 0;
  _BX = s;
  LSL_EAX_EBX
  POP_BX
  return _EAX;
#elif defined (__DPMI16__)
  _BX = s;
  LSL_EAX_EBX
  return _AX;
#else
  return 0xFFFFL;  // real mode segments are always 64k
#endif
}

#ifdef __DPMI32__
int IsValidWin32 (bits32 o)
//
// Querry the Win32 subsystem accessability of memory
//
// Returns:
//   0 - no read or write access
//   1 - read only access
//   2 - read or write access
//
{
  int numCopied;
  bits08 value;
  ReadProcessMemory (UIObject::hPIDUser, (LPVOID)o, (LPVOID)&value, 1, (LPDWORD)&numCopied);
  if (numCopied != 1)
    return 0;  // no read or write access

  return 1;
}
#endif

int UIDmp::IsValid (bits16 s, bits32 o)
{
  if (IsValidSel (s))
  {
    if (o > GetLimit (s))
      return 0;
    else
    #ifdef __DPMI32__
      if (des.allocated == 0)
        return IsValidWin32 (o);
    #endif
    return 1;
  }
  return 0;
}

#pragma warn -rch
#pragma argsused
int IsValidSel (bits16 s)
{
#if defined (__DPMI32__)
  PUSH_BX
  _EBX = 0;
  _BX = s;
  VERR_BX
  POP_BX
#elif defined (__DPMI16__)
  _BX = s;
  VERR_BX
#else
  return 1;   // segments are always valid in real mode.
#endif
  // gets here only if in prot mode (16 or 32)
  if (_FLAGS & 64)  // if zero flag is set from verr then it's valid
    return 1;
  return 0;
}
#pragma argsused
bits08 UIDmp::GetFarByte (bits16 s, unsigned int o)
{
#ifdef __DPMI32__
  int numCopied;
  bits08 value;
  if (!des.allocated)
  {
    ReadProcessMemory (UIObject::hPIDUser, (LPVOID)((bits32)o), (LPVOID)&value, 1, (LPDWORD)&numCopied);
    if (numCopied != 1)
      assert (0);  // we should already have access to this, so if it
                   // fails, then a terrible thing has happened!
    else
      _AL = value;
  }
  else
  {
    PUSH_BX
    PUSH_ES
    _ES = des.index;
    _BX = o;
    MOV_AL_ES_BX
    POP_ES
    POP_BX
  }
#else
  _ES = s;
  _BX = o;
  asm mov al, byte ptr es:[bx]
#endif
  return _AL;
}

#pragma argsused
int UIDmp::Update (int dsp)
{
  UIRect r = rWindow;
  int j,k;
  int width = r.Width();
  char buf[80], *p, c;
  bits32 off = curOfs;

  if (width > 73)
    dmpWidth = 16;
  else
    dmpWidth = 8;

  for (j=0;j<r.Height();j++)
  {
    p = buf;
    sprintf (p,
      #if defined (__DPMI32__)
        ":%08lX ",
      #else
        ":%04X ",
      #endif
      (unsigned int) off);
    p = &p[strlen(p)];
    for (k=0;k<dmpWidth;k++)
    {
    #if !defined(__DPMI32__)
      if (IsValid (curSel, ((off+k)&0x0FFFFL)))  // force segment wrap around
    #else
      if (IsValid (curSel, off+k))
    #endif
        sprintf (p, "%02X ", GetFarByte (curSel, (unsigned int)off+k));
      else
        strcpy (p, "   ");
      p = &p[strlen(p)];
    }
    for (k=0;k<dmpWidth;k++)
    {
    #if !defined(__DPMI32__)
      if (IsValid (curSel, ((off+k)&0x0FFFFL)))  // force segment wrap around
    #else
      if (IsValid (curSel, off+k))
    #endif
        c = GetFarByte (curSel, (unsigned int)off+k);
      else
        c = ' ';
      if (!isprint (c))
        c = 249;
      sprintf (p, "%c", c);
      p = &p[strlen(p)];
    }
    off += dmpWidth;
    buf[width] = 0;
    gotoxy (1,j+1);
    cputs (buf);
  }
  return 1;
}

int ishexdigit (unsigned char b)
{
  if (isdigit (b))
    return 1;

  switch (b)
  {
    case 'A':
    case 'B':
    case 'C':
    case 'D':
    case 'E':
    case 'F':
    case 'a':
    case 'b':
    case 'c':
    case 'd':
    case 'e':
    case 'f':
      return 1;
  }
  return 0;
}

int UIDmp::ParseSearch (char *s, bits08 *bytes)
{
  char *p, *q;
  int len = 0;
  int quote = 0;
  unsigned char b;
  p = s;

  while (*p)
  {
    if (*p == '\n')
      return len;

    if (*p == ' ')
    {
      p++;
      continue;
    }

    if (*p == '\'' || *p == '\"')
    {
      quote = !quote;
      p++;
      continue;
    }

    if (quote)
    {
      bytes[len++] = *p;
      if (len >= 16)
        return 16;
      p++;
    }
    else
    {
      b = (unsigned char) strtoul (p, NULL, 16);
      bytes[len++] = b;
      if (len >= 16)
        return 16;

      while (*p && ishexdigit (*p)) p++;
    }
  }
  return len;
}

bits32 UIDmp::Search (char *srch, bits32 start)
{
  bits08 bytes[16];
  int leng = 0;
  int found = 0;
  static bits32 sr, lim;
  bits08 c;
  int counter = 0;

  leng = ParseSearch (srch, bytes);
  lim = GetLimit (curSel);

  for ( sr = start; sr < lim; sr++)
  {
    if ((++counter % 500) == 0)
    {
      spin.Spin (1);
      counter = 0;
      if (kbhit ())          // only check the keyboard once in a while
        if (getch () == K_ESC)
          break;
    }

    if (IsValid (curSel, sr))
    {
      c = GetFarByte (curSel, sr);
      if (c == bytes[found])
      {
        found++;
        if (found >= leng)
        {
          srchOfs = (sr - leng) + 1;
          srchSel = curSel;
          spin.Restore ();
          return srchOfs;
        }
      }
      else
        found = 0;
    }
    else
      found = 0;
  }

  spin.Restore ();
  UIMsgBox (srch, " No further matches found ", 2, 60);
  srchOfs = 0;
  srchSel = 0;
  return start;
}

void UIDmp::begin (void)
{
  char *q, *p;
  char dest[100];
  static char srch[41] = {0};
  int out = 0, l;
  char buf[40];
  bits32 tmpOfs = 0;
  UIInputLong *il = 0;
  UIInputShort *is = 0;
  UIInputLine *iln = 0;

  SetLineType (1);
  DrawBox (uiFlags, YELLOW|_BLUE, WHITE|_BLUE);

  while (!out)
  {
    while (kbhit ()) getch ();  // suck up any extra keys

    switch (getch())
    {
      case K_F1:
        {
          UIMsgBox (
          " <Esc>         - exit to previous window\r\n"
          " <F5>          - zooms the screen\r\n"
          " <Ctrl-F5>     - move/re-size the window\r\n"
          " <Up><Down>,   \r\n"
          " <PgUp><PgDwn> - moves offset \r\n"
          " <Ctrl-PgUp>,  \r\n"
          " <Ctrl-PgDwn>  - adjusts offset by +- 0x1000\r\n"
          " <Ctrl-S>      - enter a new selector\r\n"
          " <Ctrl-G>      - go to a new offset\r\n"
          " <Ctrl-N>      - search for bytes or a 'string'\r\n"
          " <Ctrl-L>      - search next\r\n"
          " <Ctrl-Home>   - go to offset 0\r\n"
          #ifdef __DPMI32__
          " <Ctrl-End>    - go to offset 0x80000000\r\n"
          #else
          " <Ctrl-End>    - go to offset 0x8000\r\n"
          #endif
          ,  " Dump Window Help " ,13, 60);
        }
        break;

      case K_CTRL_HOME:
        curOfs = 0;
        Redraw (0);
        break;

      case K_CTRL_END:
        #ifdef __DPMI32__
          curOfs = 0x80000000U;
        #else
          curOfs = 0x8000U;
        #endif
        Redraw (0);
        break;

      case K_CTRL_F4:       // undocumented evaluate box

        iln = new UIInputLine (*this, 50, " Enter expression to evaluate ");
        p = iln->begin ();
        if (!p)
        {
          delete iln;
          break;
        }
        else
        {
          EvalExpr (p, dest);
          UIMsgBox (dest, p, 2,70);
        }
        delete iln;
        break;

      case K_CTRL_L:
        if (!srch[0])
        {
          iln = new UIInputLine (*this, 40, " Enter search bytes ");
          p = iln->begin ();
          if (p)
            strcpy (srch, p);
          else
          {
            delete iln;
            break;
          }
          delete iln;
          srchOfs = 0;
          srchSel = 0;
        }
        if ((srchSel == curSel) && (srchOfs == curOfs))
        {
          tmpOfs = Search (srch, curOfs+1);
          if (tmpOfs == curOfs+1)
            break;
          curOfs = tmpOfs;
        }
        else
          curOfs = Search (srch, curOfs);
        Redraw (0);
        break;

      case K_CTRL_N:
        iln = new UIInputLine (*this, 40, " Enter search bytes ");
        p = iln->begin ();
        if (p)
          strcpy (srch, p);
        else
        {
          delete iln;
          break;
        }
        delete iln;
        srchOfs = 0;
        srchSel = 0;
        curOfs = Search (srch, curOfs);
        Redraw (0);
        break;

      case K_UP:
        curOfs -= dmpWidth;
        #ifndef __DPMI32__
          curOfs &= 0xFFFF;
        #endif
        Redraw (0);
        break;

      case K_DOWN:
        curOfs += dmpWidth;
        #ifndef __DPMI32__
          curOfs &= 0xFFFF;
        #endif

        Redraw (0);
        break;

      case K_PGUP:
        curOfs -= dmpWidth * rWindow.Height();
        #ifndef __DPMI32__
          curOfs &= 0xFFFF;
        #endif

        Redraw (0);
        break;

      case K_PGDOWN:
        curOfs += dmpWidth * rWindow.Height();
        #ifndef __DPMI32__
          curOfs &= 0xFFFF;
        #endif
        Redraw (0);
        break;

      case K_CTRL_PGUP:
        curOfs -= 0x1000;
        #ifndef __DPMI32__
          curOfs &= 0xFFFF;
        #endif
        Redraw (0);
        break;

      case K_CTRL_PGDOWN:
        curOfs += 0x1000;
        #ifndef __DPMI32__
          curOfs &= 0xFFFF;
        #endif
        Redraw (0);
        break;

      case K_CTRL_G:
        il = new UIInputLong (*this, " Enter a new offset ");
        curOfs = il->Value (curOfs);
        delete il;
        Redraw (0);
        break;

      case K_CTRL_S:
        is = new UIInputShort (*this, " Enter a new selector ");
        curSel = is->Value (curSel);
        delete is;
        DoNewSelector (curSel);
        Redraw (1);
        break;

      case K_ENTER:
        break;

      case K_ESC:
        out = 1;
        break;
    }
  }
}
