/*
  UIWIN.CPP
  Main source file for class UIObject, UIWindow, and mouse stuff


  Copyright (c) 1994 by Borland International
  All Rights Reserved

  This source code is hereby released into the public domain.
  Written by Jeffrey J. Peters

*/

#include <dos.h>
#include "uiclass.h"

int MouseEvent = 0;
bits16 mousex,mousey;

UIObject *UIObject::curObj = 0;
bits16 UIObject::screenSel = 0;

#ifdef __DPMI32__
HANDLE UIObject::hPIDUser = (HANDLE)0;
#endif

void UIObject::LowLevelIdle (void)
{
  bits16 mx, my, button, b;
#ifdef __DPMI32__
  _EAX = 0;
#endif
  _AX = 3;
  geninterrupt (0x33);

  mx = _CX;
  my = _DX;
  button = _BX;

  if (button)
  {
    MouseEvent = 1;
    b = button;
    while (b == button)
    {
      #ifdef __DPMI32__
        _EAX = 0;
      #endif
        _AX = 3;
        geninterrupt (0x33);

        mx = _CX;
        my = _DX;
        button = _BX;
    }
    mousex = (mx / 8) + 1;
    mousey = (my / 8) + 1;
  }
}

void UIObject::Mouse (void)
{
}

void UIObject::ProcessMouse (void)
{
}

int UIObject::BIOSKeyWaiting (void)
{
  int ret = 1;

  LowLevelIdle ();

  #ifdef __DPMI32__
    _EAX = 0x0100;
  #else
    _AX = 0x0100;
  #endif
    geninterrupt (0x16);
  if (_FLAGS & 64)
    ret = 0;

  if (MouseEvent == 1)
  {
    if (IsContained (UIPoint(mousex,mousey)))
      ret = 1;
    else
      MouseEvent = 0;
  }
  return ret;
}

bits16 UIObject::GetBIOSKey (void)
{
  while (!BIOSKeyWaiting ()) ;
  if (MouseEvent == 1)
  {
    MouseEvent = 0;
    return K_ESC;
  }

  #ifdef __DPMI32__
    _EAX = 0x0000;
  #else
    _AX = 0x0000;
  #endif
    geninterrupt (0x16);
  return _AX;
}

int UIRect::Contained (UIPoint &p)
{
  if ((p.x >= ul.x) && (p.x <= lr.x))
    if ((p.y >= ul.y) && (p.y <= lr.y))
      return 1;
  return 0;
}

char *newStr (char *s)
{
  bits16 sz;
  char *p;

  assert (s);
  sz = strlen (s);
  p = new char [sz+1];
  assert (p);
  strcpy (p, s);
  return p;
}

char UIWindow::lineType[2][6] =
{
  "ĳڿ",
  "ͺɻ"
};

char *UIWindow::GetLineStr (int set, int chr)
{
  static char buf[2];

  if (set < 2 && chr < 6)
  {
    buf[1] = 0;
    buf[0] = lineType[set][chr];
    return buf;
  }
  return 0;
}

int UIWindow::SetLineType (int x)
{
  int old;

  if (x > 2 || x < 0)
    x = 0;
  old = lineSet;
  lineSet = x;
  return old;
}

UIWindow::UIWindow (UIRect r, char *tstr) : UIObject (),
                                            screenSave (0), rFrame (0,0,0,0),
                                            rWindow (0,0,0,0), bgfgBorder (0),
                                            bgfgScreen (0), uiFlags (0),
                                            lineSet (0), shown (0), zoomed (0),
                                            rOrig (0,0,0,0)
{
  _wscroll = 0;
  _setcursortype (_NOCURSOR);

  if (tstr)
    title = newStr (tstr);
  else
    title = newStr ("Untitled");

  rFrame = r;
  rWindow = r;
  rWindow.lr -= UIPoint (2,1);
  rWindow += UIRect (1, 1, -1, -1);
  UISave ();
}

UIWindow::UIWindow () : UIObject (), screenSave (0), rFrame (0,0,0,0),
                        rWindow (0,0,0,0), bgfgBorder (0),
                        bgfgScreen (0), uiFlags (0),
                        lineSet (0), title (0), memSize (0), shown (0),
                        zoomed (0), rOrig (0,0,0,0)
{
}

UIWindow::~UIWindow ()
{
  if (screenSave)
  {
    UIRestore ();
  }
  if (!(uiFlags & UIFNoCursorFix))
    if (uiFlags & UIFShowCursor)
      _setcursortype (_NOCURSOR);
    else
      _setcursortype (_NORMALCURSOR);
  if (title)
    delete title;
}

#if 0  // used only for debugging
void UIWindow::MemLock ()
{
  if (!screenSave)
    return;

  TDescriptor des (FP_SEG(screenSave));
  des.SetRights (des.rights & (~0x80)); // mark the segment as not present
}

void UIWindow::MemUnlock ()
{
  if (!screenSave)
    return;

  TDescriptor des (FP_SEG(screenSave));
  des.SetRights (des.rights | 0x80);  // mark the segment as present
}
#else
void UIWindow::MemLock () {}
void UIWindow::MemUnlock () {}
#endif

void UIWindow::UISave ()
{
  UIRect &r = rFrame;
  if (screenSave)
    assert (0);
  memSize = r.Area()*2 + 2;
  screenSave = new char [memSize];
  MouseHide();
  gettext (r.ul.x, r.ul.y, r.lr.x, r.lr.y, screenSave);
  MemLock ();
  MouseShow();
}

void UIWindow::UIRestore ()
{
  UIRect &r = rFrame;
  MouseHide();
  MemUnlock ();
  puttext (r.ul.x, r.ul.y, r.lr.x, r.lr.y, screenSave);
  delete [] screenSave;
  screenSave = 0;
  MouseShow();
}

bits16 GetCharAt (bits16 x, bits16 y)
{
  struct text_info ti;
  bits16 buf[10] = {0};
  gettextinfo (&ti);
  MouseHide();
  gettext (ti.winleft+x-1, ti.wintop+y-1, ti.winleft+x-1, ti.wintop+y-1, buf);
  MouseShow();
  return buf[0];
}

void PutCharAt (bits16 x, bits16 y, bits16 buf)
{
  struct text_info ti;
  gettextinfo (&ti);
  MouseHide();
  puttext (ti.winleft+x-1, ti.wintop+y-1, ti.winleft+x-1, ti.wintop+y-1, &buf);
  MouseShow();
}

void UIWindow::ShowTitle (bits08 bgfgTitle)
{
  bits16 l;
  char *p = title;
  UIRect r;

  r = rWindow;
  window (r.ul.x, r.ul.y, r.lr.x, r.lr.y);

  l = strlen (p);
  if (l >= (r.Width()-4))
    p = "[.]";

  l = l / 2;
  l = (r.Width() /2) - l;
  gotoxy (l,1);
  textattr (bgfgTitle);
  MouseHide();
  cprintf ("[%s]", p);
  MouseShow();
}

#pragma argsused
void UIWindow::DrawBox (bits16 flags, bits08 bgfgS, bits08 bgfgB)
{
  UIRect r;
  bits16 x,y, j;
  static char linebuf[80];

  bgfgScreen = bgfgS;
  bgfgBorder = bgfgB;
  uiFlags    = flags;

  MouseHide ();

  if (!(flags & UIFNoShadow))
  {
    r = rFrame;
    window (r.ul.x, r.ul.y, r.lr.x, r.lr.y);
    for (y = 2; y < r.Height(); y++)
    {
      j = GetCharAt (r.Width(),y);
      j = (j & 0x00FF) | ((DARKGRAY|_BLACK)  << (bits16)8);
      PutCharAt (r.Width(),y, j);

      j = GetCharAt (r.Width()-1,y);
      j = (j & 0x00FF) | ((DARKGRAY|_BLACK)  << (bits16)8);
      PutCharAt (r.Width()-1,y, j);
    }

    for (x = 3; x <= r.Width(); x++)
    {
      gotoxy (x, r.Height());
      j = GetCharAt (x, r.Height());
      j = (j & 0x00FF) | ((DARKGRAY|_BLACK)  << (bits16)8);
      PutCharAt (x, r.Height(), j);
    }
  }

  rWindow -= UIRect (1, 1, -1, -1);
  r = rWindow;
  window (r.ul.x, r.ul.y, r.lr.x, r.lr.y);
  if (!(flags & UIFNoClearScreen))
  {
    textattr (bgfgScreen);
    clrscr ();
  }

  textattr (bgfgBorder);
  for (y = 1; y < r.Height(); y++)
  {
    gotoxy (1, y);
    cputs (GetLineStr (lineSet,1));
    gotoxy (r.Width(), y);
    cputs (GetLineStr (lineSet,1));
  }

  memset (linebuf, GetLineStr (lineSet,0)[0],r.Width());
  linebuf[r.Width()] = 0;
  gotoxy (1, 1);
  cputs (linebuf);
  gotoxy (1, r.Height());
  cputs (linebuf);

  gotoxy (1,1);
  cputs (GetLineStr (lineSet,2));
  gotoxy (r.Width(), 1);
  cputs (GetLineStr (lineSet,3));
  gotoxy (1,r.Height());
  cputs (GetLineStr (lineSet,5));
  gotoxy (r.Width(),r.Height());
  cputs (GetLineStr (lineSet,4));
  ShowTitle (bgfgBorder);
  rWindow += UIRect (1, 1, -1, -1);

  shown = 1;
  CallUpdate (1);
  MouseShow ();
}

void UIWindow::Zoom (int inOut)
{
  if (uiFlags & UIFNoZoom)
    return;

  if (zoomed == inOut)
    return;

  zoomed = !(!inOut) ;  // force zoomed to true or false

  if (inOut)
  {
    if (rOrig == UIRect (0,0,0,0))
      rOrig = rFrame;

    UIRect r = GetMaxRect ();
    if ((rFrame == r) && (r == rReg))
      rReg = rOrig;
    else
      rReg = rFrame;
    MoveTo (UIRect (r));
  }
  else
  {
    UIRect r = rFrame;
    MoveTo (rReg);
    rReg = r;
  }
}

void UIWindow::SlideOver (bits16 dir)
{
  UIRect r = rFrame;
  // dir = 0 for no movement
  //     = 1 for left
  //     = 2 for right
  //     = 3 for up
  //     = 4 for down
  if (dir)
  {
    switch (dir)
    {
      case 1:
        r -= UIRect (1,0,1,0);
        break;

      case 2:
        r += UIRect (1,0,1,0);
        break;

      case 3:
        r -= UIRect (0,1,0,1);
        break;

      case 4:
        r += UIRect (0,1,0,1);
        break;
    }

    if ((r.lr.y > 25) || (r.lr.x > 80))
      return;

    if ((r.ul.y < 1) || (r.ul.x < 1))
      return;

    UIRestore ();
    rFrame = r;
    rWindow = r;
    rWindow.lr -= UIPoint (2,1);
    rWindow += UIRect (1, 1, -1, -1);
    UISave ();
    if (shown)
      DrawBox (uiFlags, bgfgScreen, bgfgBorder);
  }
}

int Sign (int x)
// returns -1 if less than zero
//          0 if equal to zero
//         +1 if greater than zero
{
  if (x == 0)
    return 0;

  if (x < 0)
    return -1;

  return 1;
}

void UIWindow::Stretch (int dir)
{
  UIRect r = rFrame;
  // dir = 0 for no movement
  //     = 1 for left
  //     = 2 for right
  //     = 3 for up
  //     = 4 for down
  //     the sign determines increase or decrease

  if (uiFlags & UIFNoStretch)
    return ;

  if (zoomed)
    return ;

  if (dir)
  {
    switch (abs (dir))
    {
      case 1:
        r -= UIRect (1*Sign(dir),0,0,0);
        break;

      case 2:
        r += UIRect (0,0,1*Sign(dir),0);
        break;

      case 3:
        r -= UIRect (0,1*Sign(dir),0,0);
        break;

      case 4:
        r += UIRect (0,0,0,1*Sign(dir));
        break;
    }

    if ((r.lr.y > 25) || (r.lr.x > 80))
      return;

    if ((r.ul.y < 1) || (r.ul.x < 1))
      return;

    if (r.Width () < MinWidth ())
      return;

    if (r.Height () < MinHeight ())
      return;

    UIRestore ();
    rFrame = r;
    rWindow = r;
    rWindow.lr -= UIPoint (2,1);
    rWindow += UIRect (1, 1, -1, -1);
    UISave ();
    if (shown)
      DrawBox (uiFlags, bgfgScreen, bgfgBorder);
  }
}

int UIWindow::MoveTo (UIRect r)
{
  if ((r.lr.y > 25) || (r.lr.x > 80))
    return 0;

  if ((r.ul.y < 1) || (r.ul.x < 1))
    return 0;

  UIRestore ();
  rFrame = r;
  rWindow = r;
  rWindow.lr -= UIPoint (2,1);
  rWindow += UIRect (1, 1, -1, -1);
  UISave ();
  if (shown)
    DrawBox (uiFlags, bgfgScreen, bgfgBorder);
  return 1;
}

void UIWindow::Redraw (int flag)
{
  if (flag)
  {
    UIRestore ();
    UISave ();
    DrawBox (uiFlags, bgfgScreen, bgfgBorder);
  }
  else
  {
    CallUpdate (flag);
  }
}

void UIWindow::CallUpdate (int dsp)
{
  UIRect &r = rWindow;
  if (uiFlags & UIFNoScroll)
    _wscroll = 0;
  else
    _wscroll = 1;

  window (r.ul.x, r.ul.y, r.lr.x, r.lr.y);
  textattr (bgfgScreen);
  if (uiFlags & UIFShowCursor)
    _setcursortype (_NORMALCURSOR);

  Update (dsp);
  _wscroll = 0;
}

#pragma argsused
int UIWindow::Update (int dsp)
{
  return 1;
}

bits16 UIWindow::getch ()
{
  bits16 b;

  b = UIObject::GetBIOSKey ();

  switch (b)
  {
    case K_ALT_F3:    // close window
      b = K_ESC;
      break;

    case K_F5:        // zoom
      Zoom (!IsZoomed ());
      b = getch ();   // recursion is intended here
      break;

    case K_CTRL_F5:   // move
      if (IsZoomed ())
      {
        b = getch ();  // and here too.
        break;
      }
      UIRect r;
      UIMove *mv = new UIMove (*this);
      if (mv->Move () == 0)
      {
        delete mv;
      }
      else
      {
        r = mv->GetFrame();
        if (mv->IsZoomed ())
        {
          delete mv;
          Zoom (1);
        }
        else
        {
          delete mv;
          MoveTo (r);
          Redraw (0);
        }
      }
      b = getch ();     // recursion is intended here
      break;
  }
  return b;
}

int UIWindow::kbhit ()
{
  int i;
  i = UIObject::BIOSKeyWaiting();
  return i;
}

void MouseShow (void)
{
  #ifdef __DPMI32__
    _EAX = 0;
  #endif
    _AX = 1;
    geninterrupt (0x33);
}

void MouseHide (void)
{
  #ifdef __DPMI32__
    _EAX = 0;
  #endif
    _AX = 2;
    geninterrupt (0x33);
}